from __future__ import annotations

import contextlib
import gzip
from pathlib import Path
from typing import IO, TYPE_CHECKING, BinaryIO

import pytest

if TYPE_CHECKING:
    from collections.abc import Iterator


def absolute_path(filename: str) -> Path:
    return Path(__file__).parent / filename


def open_file(name: str, mode: str = "rb") -> Iterator[IO]:
    with absolute_path(name).open(mode) as f:
        yield f


def open_file_gz(name: str, mode: str = "rb") -> Iterator[gzip.GzipFile]:
    with gzip.GzipFile(absolute_path(name), mode) as f:
        yield f


def open_files_gz(names: list[str], mode: str = "rb") -> Iterator[list[gzip.GzipFile]]:
    with contextlib.ExitStack() as stack:
        yield [stack.enter_context(gzip.GzipFile(absolute_path(name), mode)) for name in names]


@pytest.fixture
def btrfs_default() -> Iterator[BinaryIO]:
    yield from open_file_gz("_data/btrfs-default.bin.gz")


@pytest.fixture
def btrfs_sparse() -> Iterator[BinaryIO]:
    yield from open_file_gz("_data/btrfs-sparse.bin.gz")


@pytest.fixture
def btrfs_subvolume() -> Iterator[BinaryIO]:
    yield from open_file_gz("_data/btrfs-subvolume.bin.gz")


@pytest.fixture
def btrfs_subvolume_nested() -> Iterator[BinaryIO]:
    yield from open_file_gz("_data/btrfs-subvolume-nested.bin.gz")


@pytest.fixture
def btrfs_subvolume_custom_default() -> Iterator[BinaryIO]:
    yield from open_file_gz("_data/btrfs-subvolume-custom-default.bin.gz")


@pytest.fixture
def btrfs_subvolume_snapshot() -> Iterator[BinaryIO]:
    yield from open_file_gz("_data/btrfs-subvolume-snapshot.bin.gz")


@pytest.fixture
def btrfs_compression() -> Iterator[BinaryIO]:
    yield from open_file_gz("_data/btrfs-compression.bin.gz")


@pytest.fixture
def btrfs_profile_dup() -> Iterator[list[BinaryIO]]:
    yield from open_files_gz(["_data/btrfs-dup-1.bin.gz", "_data/btrfs-dup-2.bin.gz"])


@pytest.fixture
def btrfs_profile_raid0() -> Iterator[list[BinaryIO]]:
    yield from open_files_gz(["_data/btrfs-raid0-1.bin.gz", "_data/btrfs-raid0-2.bin.gz"])


@pytest.fixture
def btrfs_profile_raid1() -> Iterator[list[BinaryIO]]:
    yield from open_files_gz(["_data/btrfs-raid1-1.bin.gz", "_data/btrfs-raid1-2.bin.gz"])


@pytest.fixture
def btrfs_profile_raid1c3() -> Iterator[list[BinaryIO]]:
    yield from open_files_gz(
        [
            "_data/btrfs-raid1c3-1.bin.gz",
            "_data/btrfs-raid1c3-2.bin.gz",
            "_data/btrfs-raid1c3-3.bin.gz",
        ]
    )


@pytest.fixture
def btrfs_profile_raid1c4() -> Iterator[list[BinaryIO]]:
    yield from open_files_gz(
        [
            "_data/btrfs-raid1c4-1.bin.gz",
            "_data/btrfs-raid1c4-2.bin.gz",
            "_data/btrfs-raid1c4-3.bin.gz",
            "_data/btrfs-raid1c4-4.bin.gz",
        ]
    )


@pytest.fixture
def btrfs_profile_raid5() -> Iterator[list[BinaryIO]]:
    yield from open_files_gz(["_data/btrfs-raid5-1.bin.gz", "_data/btrfs-raid5-2.bin.gz"])


@pytest.fixture
def btrfs_profile_raid6() -> Iterator[list[BinaryIO]]:
    yield from open_files_gz(["_data/btrfs-raid6-1.bin.gz", "_data/btrfs-raid6-2.bin.gz", "_data/btrfs-raid6-3.bin.gz"])


@pytest.fixture
def btrfs_profile_raid10() -> Iterator[list[BinaryIO]]:
    yield from open_files_gz(["_data/btrfs-raid10-1.bin.gz", "_data/btrfs-raid10-2.bin.gz"])



from __future__ import annotations

from functools import lru_cache
from typing import TYPE_CHECKING, Literal

from dissect.btrfs.c_btrfs import c_btrfs

if TYPE_CHECKING:
    from collections.abc import Iterator

    from dissect.btrfs.btrfs import Btrfs


class BTree:
    """Represent a Btrfs B-tree.

    One of ``root_item`` or ``root_offset`` must be given.

    Args:
        btrfs: The filesystem this B-tree belongs to.
        root_item: Optional ``btrfs_root_item`` to open the B-tree from.
        root_offset: Optional offset to open the B-tree from.
    """

    def __init__(self, btrfs: Btrfs, root_item: c_btrfs.btrfs_root_item | None = None, root_offset: int | None = None):
        self.btrfs = btrfs

        if not root_item and not root_offset:
            raise ValueError("Either root_item or root_offset are required")

        self.root_item = root_item
        self.root_offset = root_offset or root_item.bytenr

        self._read_node = lru_cache(8192)(self._read_node)

    def cursor(self) -> Cursor:
        """Return a new cursor into the B-tree."""
        return Cursor(self)

    def find(
        self, objectid: int | None = None, type: int | None = None, offset: int | None = None
    ) -> tuple[c_btrfs.btrfs_item, memoryview]:
        """Search for a single item in the B-tree.

        Args:
            objectid: Optional object ID to search for.
            type: Optional type to search for.
            offset: Optional offset to search for.
        """
        cursor = self.cursor()
        if not cursor.search(objectid, type, offset) or _cmp_key(cursor.item().key, objectid, type, offset) != 0:
            raise KeyError(f"Can't find item with key ({objectid}, {type}, {offset})")

        return cursor.get()

    def _read_node(self, address: int) -> tuple[memoryview, c_btrfs.btrfs_header]:
        """Helper method for reading a node in the B-tree.

        Args:
            address: The address of the node to read.
        """
        node = memoryview(self.btrfs._read_node(address))
        header = c_btrfs.btrfs_header(node)

        return node, header


class Cursor:
    """A basic cursor implementation for interacting with a :class:`BTree`.

    Args:
        btree: The :class:`BTree` to open a cursor on.
    """

    def __init__(self, btree: BTree):
        self.btree = btree

        self._node = None
        self._header = None
        self._keys = {}
        self._items = {}
        self._index = None
        self._path = []
        self.reset()

    def reset(self) -> None:
        """Reset the cursor."""
        self._node = None
        self._header = None
        self._keys = {}
        self._items = {}
        self._index = None
        self._path = []
        self.push(self.btree.root_offset)

    def push(self, address: int, initial_index: int = 0) -> None:
        """Push the cursor down a node.

        Args:
            address: The address of the node to push down to.
            initial_index: The initial index to seek to in the new node. ``-1`` means the last item.
        """
        self._path.append((self._node, self._header, self._keys, self._items, self._index))

        self._node, self._header = self.btree._read_node(address)
        self._keys = {}
        self._items = {}
        self._index = self._header.nritems if initial_index == -1 else initial_index

    def pop(self) -> None:
        """Pop up a node."""
        self._node, self._header, self._keys, self._items, self._index = self._path.pop()

    def _read_key(self, index: int) -> c_btrfs.btrfs_disk_key:
        """Read a key at the specified index.

        Args:
            index: The index of the key to read.
        """
        if key := self._keys.get(index):
            return key

        struct = c_btrfs.btrfs_key_ptr if self._header.level else c_btrfs.btrfs_item
        offset = len(c_btrfs.btrfs_header) + (len(struct) * index)
        key = c_btrfs.btrfs_disk_key(self._node[offset : offset + len(c_btrfs.btrfs_disk_key)])
        self._keys[index] = key

        return key

    def _read_item(self, index: int) -> c_btrfs.btrfs_key_ptr | c_btrfs.btrfs_item:
        """Read an item at the specified index.

        Args:
            index: The index of the item to read.
        """
        if item := self._items.get(index):
            return item

        struct = c_btrfs.btrfs_key_ptr if self._header.level else c_btrfs.btrfs_item
        offset = len(c_btrfs.btrfs_header) + (len(struct) * index)
        item = struct(self._node[offset : offset + len(struct)])
        self._items[index] = item

        if index not in self._keys:
            self._keys[index] = item.key

        return item

    def next(self) -> None:
        """Traverse to the next leaf item.

        Will move up and down the tree to find the next leaf item.
        """
        self.next_node()
        while self._header.level:
            self.push(self.item().blockptr)

    def next_node(self) -> None:
        """Traverse to the next node.

        Will move up and down the tree to find the next node.
        """
        while self._index is not None and self._index + 1 >= self._header.nritems:
            self.pop()

        if self._index is None:
            raise ValueError("Reached end")

        self._index += 1
        if self._header.level:
            self.push(self.item().blockptr)

    def prev(self) -> None:
        """Traverse to the previous leaf item.

        Will move up and down the tree to find the previous leaf item.
        """
        self.prev_node()
        while self._header.level:
            self.push(self.item().blockptr, -1)

    def prev_node(self) -> None:
        """Traverse to the previous leaf item.

        Will move up and down the tree to find the previous leaf item.
        """
        while self._index is not None and self._index - 1 < 0:
            self.pop()

        if self._index is None:
            raise ValueError("Reached start")

        self._index -= 1
        if self._header.level:
            self.push(self.item().blockptr, -1)

    def first(self) -> None:
        """Move the cursor to the first leaf item."""
        self.reset()
        while self._header.level:
            self.push(self.item().blockptr)

    def last(self) -> None:
        """Move the cursor to the last leaf item."""
        self.reset()
        self._index = self._header.nritems - 1
        while self._header.level:
            self.push(self.item().blockptr, -1)

    def get(self) -> tuple[c_btrfs.btrfs_item, memoryview]:
        """Retrieve the leaf item and the associated data at the current cursor position.

        Cursor must be positioned at a leaf item.
        """
        return self.item(), self.data()

    def item(self) -> c_btrfs.btrfs_key_ptr | c_btrfs.btrfs_item:
        """Retrieve a leaf or branch item.

        Cursor can be positioned at a branch or leaf item.
        """
        if self._index is None:
            raise ValueError("Cursor not set")

        return self._read_item(self._index)

    def items(self) -> Iterator[c_btrfs.btrfs_key_ptr | c_btrfs.btrfs_item]:
        """Iterate over all items in the current node."""
        for i in range(self._header.nritems):
            yield self._read_item(i)

    def data(self) -> memoryview:
        """Return the associated data of the current leaf item."""
        if self._index is None:
            raise ValueError("Cursor not set")

        if self._header.level:
            raise ValueError("Cursor not set to a leaf")

        header_size = len(c_btrfs.btrfs_header)
        item = self.item()

        offset = header_size + item.offset
        return self._node[offset : offset + item.size]

    def iter(
        self,
        objectid: int | None = None,
        type: int | None = None,
        offset: int | None = None,
        ignore_offset: bool = False,
    ) -> Iterator[tuple[c_btrfs.btrfs_item, memoryview]]:
        """Search and iterate the B-tree for the specified key.

        Stop iterating if the current item no longer matches the given parameters.

        Args:
            objectid: Optional object ID to search for.
            type: Optional type to search for.
            offset: Optional offset to search for.
            ignore_offset: Only use the ``offset`` argument for the initial positioning of the cursor,
                but ignore it for future iterations. This is useful for e.g. iterating file extents.
        """
        if not self.search(objectid, type, offset):
            return

        while _cmp_key(self.item().key, objectid, type, None if ignore_offset else offset) == 0:
            yield self.get()

            try:
                self.next()
            except ValueError:
                return

    def walk(
        self,
        objectid: int | None = None,
        type: int | None = None,
        offset: int | None = None,
    ) -> Iterator[tuple[c_btrfs.btrfs_item, memoryview]]:
        """Walk all leaf items of the B-tree and yield all matching leafs.

        Args:
            objectid: Optional object ID to yield items of.
            type: Optional type to yield items of.
            offset: Optional offset to yield items of.
        """
        self.first()

        while True:
            if _cmp_key(self.item().key, objectid, type, offset) == 0:
                yield self.get()

            try:
                self.next()
            except ValueError:
                return

    def search(self, objectid: int | None = None, type: int | None = None, offset: int | None = None) -> bool:
        """Perform a binary search on the current node for the key with the given parameters.

        Puts the cursor at the index of the matching item, or just before a "greater" item if no exact match is found.
        ``True`` is returned if this is the case, else ``False`` is returned and the cursor position is reset.

        Args:
            objectid: Optional object ID to search for.
            type: Optional type to search for.
            offset: Optional offset to search for.
        """
        while True:
            min_idx = 0
            max_idx = self._header.nritems - 1

            while min_idx != max_idx:
                test_idx = (min_idx + max_idx) // 2
                key = self._read_key(test_idx)

                result = _cmp_key(key, objectid, type, offset or 0)
                if result < 0:
                    min_idx = test_idx + 1
                else:
                    max_idx = test_idx

            result = _cmp_key(self._read_key(min_idx), objectid, type, offset)
            if self._header.level:
                if result > 0 and min_idx > 0:
                    # When we're at a node level, and the next key is larger than what we're searching for, then
                    # we must traverse down the previous node
                    min_idx -= 1
                self._index = min_idx
                self.push(self._read_item(min_idx).blockptr)
                continue

            if result >= 0:
                # Count a matching or larger key as a win
                self._index = min_idx
                return True

            if min_idx == self._header.nritems - 1:
                # Special case where we have exhausted all leaf nodes but all keys are still smaller
                # In this case, try to travel to the next node (up one level, next item, down one level)
                # Worst case we end up at a key that's larger than our search parameters.
                self._index = min_idx
                try:
                    self.next_node()
                    continue
                except ValueError:
                    self._index = None
                    return False

            break

        self._index = None
        return False


def _cmp_key(
    key: c_btrfs.btrfs_disk_key,
    objectid: int | None = None,
    type: int | None = None,
    offset: int | None = None,
) -> Literal[1, -1, 0]:
    """Compare a B-tree key on disk to a given object ID, type and offset.

    Returns -1 if the key on disk is less, 0 if the key is equal and 1 if the key is greater.
    Search parameters of ``None`` are ignored in the comparison.

    Args:
        key: The key on disk to compare.
        objectid: Optional object ID to compare against.
        type: Optional type to compare against.
        offset: Optional offset to compare against.
    """
    if objectid is not None:
        if key.objectid > objectid:
            return 1
        if key.objectid < objectid:
            return -1
    if type is not None:
        if key.type > type:
            return 1
        if key.type < type:
            return -1
    if offset is not None:
        if key.offset > offset:
            return 1
        if key.offset < offset:
            return -1
    return 0



import datetime
from typing import BinaryIO

import pytest
from dissect.util.stream import BufferedStream

from dissect.btrfs.btrfs import Btrfs
from dissect.btrfs.c_btrfs import c_btrfs
from dissect.btrfs.exceptions import Error


def assert_test_data(fs: Btrfs) -> None:
    entry = fs.get("")
    assert entry.is_dir()
    assert entry.path == ""

    entry = fs.get("path")
    assert entry.is_dir()
    assert not entry.is_file()
    assert not entry.is_symlink()
    assert not entry.is_device()
    assert not entry.is_ipc()
    assert entry.size == 4
    assert entry.path == "path"
    assert entry.uid == 1000
    assert entry.gid == 1000

    entry = fs.get("path/to/a/file.txt")
    assert entry.is_file()
    assert entry.size == 12
    assert entry.path == "path/to/a/file.txt"
    assert entry.open().read() == b"file in dir\n"

    entry = fs.get("small.txt")
    assert entry.is_file()
    assert entry.size == 29
    assert entry.path == "small.txt"
    assert entry.open().read() == b"small file content goes here\n"

    entry = fs.get("large.txt")
    assert entry.is_file()
    assert entry.size == 5242881
    assert entry.path == "large.txt"
    assert entry.open().read() == (b"a" * 5242880) + b"\n"

    entry = fs.get("link.txt")
    assert entry.is_symlink()
    assert entry.path == "link.txt"
    assert entry.link == "path/to/a/file.txt"


def test_btrfs(btrfs_default: BinaryIO) -> None:
    fs = Btrfs(btrfs_default)

    assert fs.label == "btrfs-default"
    assert str(fs.uuid) == "74387226-fa97-4f42-a276-9bb07ce5e62d"

    root = fs.get("/")
    assert list(root.listdir().keys()) == [".", "..", "path", "link.txt", "small.txt", "large.txt"]

    assert_test_data(fs)

    entry = fs.get("path")
    assert entry.mode == 0o40755
    assert entry.atime == datetime.datetime(2023, 6, 28, 3, 4, 16, tzinfo=datetime.timezone.utc)
    assert entry.ctime == datetime.datetime(2023, 6, 28, 3, 4, 12, tzinfo=datetime.timezone.utc)
    assert entry.mtime == datetime.datetime(2023, 6, 28, 3, 4, 12, tzinfo=datetime.timezone.utc)
    assert entry.otime == datetime.datetime(1970, 1, 1, 0, 0, tzinfo=datetime.timezone.utc)


def test_btrfs_subvolume(btrfs_subvolume: BinaryIO) -> None:
    fs = Btrfs(btrfs_subvolume)

    assert fs.label == "btrfs-subvolume"
    assert str(fs.uuid) == "74387226-fa97-4f42-a276-9bb07ce5e62d"

    subvol = fs.get("subvol")
    assert subvol.path == ""
    assert subvol.subvolume.path == "subvol"
    assert subvol.full_path == "subvol"
    assert list(subvol.listdir().keys()) == [".", "..", "cross-volume-link.txt", "small.txt", "large.txt", "some"]

    assert_test_data(fs)

    entry = fs.get("subvol/some/more/dirs/empty.txt")
    assert entry.is_file()
    assert entry.size == 0
    assert entry.path == "some/more/dirs/empty.txt"
    assert entry.full_path == "subvol/some/more/dirs/empty.txt"
    assert entry.open().read() == b""

    entry = fs.get("subvol/small.txt")
    assert entry.is_file()
    assert entry.size == 18
    assert entry.open().read() == b"file in subvolume\n"

    entry = fs.get("subvol/large.txt")
    assert entry.is_file()
    assert entry.size == 5242881
    assert entry.open().read() == (b"b" * 5242880) + b"\n"

    entry = fs.get("subvol/cross-volume-link.txt")
    assert entry.is_symlink()
    assert entry.link == "../link.txt"


def test_btrfs_subvolume_custom_default(btrfs_subvolume_custom_default: BinaryIO) -> None:
    fs = Btrfs(btrfs_subvolume_custom_default)

    assert fs.label == "btrfs-subvolume-custom-default"
    assert str(fs.uuid) == "74387226-fa97-4f42-a276-9bb07ce5e62d"


def test_btrfs_subvolume_nested(btrfs_subvolume_nested: BinaryIO) -> None:
    fs = Btrfs(btrfs_subvolume_nested)

    assert fs.label == "btrfs-subvolume-nested"
    assert str(fs.uuid) == "983bda53-2928-4182-a53d-041eade66473"

    subvolumes = list(fs.subvolumes())

    assert len(subvolumes) == 4
    assert sorted((subvol.objectid, subvol.path) for subvol in subvolumes) == [
        (5, ""),
        (256, "dir/volume"),
        (257, "default"),
        (258, "default/volume"),
    ]

    assert fs.find_subvolume("default/volume").objectid == 258


def test_btrfs_subvolume_snapshot(btrfs_subvolume_snapshot: BinaryIO) -> None:
    fs = Btrfs(btrfs_subvolume_snapshot)

    assert fs.label == "btrfs-subvolume-snapshot"
    assert str(fs.uuid) == "74387226-fa97-4f42-a276-9bb07ce5e62d"

    assert_test_data(fs)


def test_btrfs_compression(btrfs_compression: BinaryIO) -> None:
    fs = Btrfs(btrfs_compression)

    fh = fs.get("zlib.txt").open()
    assert len([extent for extent in fh.extents if extent.compression == c_btrfs.BTRFS_COMPRESS_ZLIB]) == 160
    assert fh.read() == (b"zlib" * 1024 * 1024 * 5) + b"\n"

    fh = fs.get("zlib_inline.txt").open()
    assert isinstance(fh, BufferedStream)
    assert fh.read() == (b"zlib" * 256) + b"\n"

    fh = fs.get("lzo.txt").open()
    assert len([extent for extent in fh.extents if extent.compression == c_btrfs.BTRFS_COMPRESS_LZO]) == 120
    assert fh.read() == b"lzo" * 1024 * 1024 * 5 + b"\n"

    fh = fs.get("lzo_inline.txt").open()
    assert isinstance(fh, BufferedStream)
    assert fh.read() == (b"lzo" * 256) + b"\n"

    fh = fs.get("zstd.txt").open()
    assert len([extent for extent in fh.extents if extent.compression == c_btrfs.BTRFS_COMPRESS_ZSTD]) == 160
    assert fh.read() == (b"zstd" * 1024 * 1024 * 5) + b"\n"

    fh = fs.get("zstd_inline.txt").open()
    assert isinstance(fh, BufferedStream)
    assert fh.read() == (b"zstd" * 256) + b"\n"


@pytest.mark.parametrize(
    "fixture",
    [
        "btrfs_profile_dup",
        "btrfs_profile_raid0",
        "btrfs_profile_raid1",
        "btrfs_profile_raid1c3",
        "btrfs_profile_raid1c4",
        "btrfs_profile_raid5",
        "btrfs_profile_raid6",
        "btrfs_profile_raid10",
    ],
)
def test_btrfs_profiles(fixture: str, request: pytest.FixtureRequest) -> None:
    fs = Btrfs(request.getfixturevalue(fixture))
    assert_test_data(fs)


@pytest.mark.parametrize(
    "fixture",
    [
        "btrfs_profile_dup",
        "btrfs_profile_raid0",
        "btrfs_profile_raid1",
        "btrfs_profile_raid1c3",
        "btrfs_profile_raid1c4",
        "btrfs_profile_raid5",
        "btrfs_profile_raid6",
        "btrfs_profile_raid10",
    ],
)
def test_btrfs_profiles_partial(fixture: str, request: pytest.FixtureRequest) -> None:
    fhs = request.getfixturevalue(fixture)
    # The btrfs_profile_dup test data just so happens to have all chunks on the second device
    fhs = fhs[:-1] if fixture == "btrfs_profile_dup" else fhs[1:]

    if fixture in ("btrfs_profile_dup", "btrfs_profile_raid0"):
        with pytest.raises(Error, match="Missing stripe disk for chunk offset .+"):
            fs = Btrfs(fhs)

    elif fixture in ("btrfs_profile_raid5", "btrfs_profile_raid6"):
        pass

    else:
        fs = Btrfs(fhs)
        assert_test_data(fs)


def test_btrfs_sparse(btrfs_sparse: BinaryIO) -> None:
    fs = Btrfs(btrfs_sparse)

    entry = fs.get("sparse_start")
    assert entry.size == 0x3C000
    assert entry.extents() == [
        (0, 0, 0, 0, 0, 40 * 4096),
        (0, 0, 0xD28000, 20 * 4096, 0, 20 * 4096),
    ]
    assert entry.open().read() == ((b"\x00" * 4096) * 40) + ((b"\x01" * 4096) * 20)

    entry = fs.get("sparse_hole")
    assert entry.size == 0x3C000
    assert entry.extents() == [
        (0, 0, 0xD00000, 20 * 4096, 0, 20 * 4096),
        (0, 0, 0, 0, 0, 20 * 4096),
        (0, 0, 0xD14000, 20 * 4096, 0, 20 * 4096),
    ]
    assert entry.open().read() == ((b"\x01" * 4096) * 20) + ((b"\x00" * 4096) * 20) + ((b"\x01" * 4096) * 20)

    entry = fs.get("sparse_end")
    assert entry.size == 0x3C000
    assert entry.extents() == [
        (0, 0, 0xD3C000, 20 * 4096, 0, 20 * 4096),
        (0, 0, 0, 0, 0, 40 * 4096),
    ]
    assert entry.open().read() == ((b"\x01" * 4096) * 20) + ((b"\x00" * 4096) * 40)

    entry = fs.get("sparse_all")
    assert entry.size == 0x500000
    assert entry.extents() == [
        (0, 0, 0, 0, 0, 1280 * 4096),
    ]
    assert entry.open().read() == (b"\x00" * 4096) * 1280

    entry = fs.get("snapshot/sparse_hole")
    assert entry.size == 0x3C000
    assert entry.extents() == [
        (0, 0, 0xD00000, 20 * 4096, 0, 10 * 4096),
        (0, 0, 0xD50000, 1 * 4096, 0, 1 * 4096),
        (0, 0, 0, 0, 0, 39 * 4096),
        (0, 0, 0xD52000, 10 * 4096, 0, 10 * 4096),
    ]
    assert entry.open().read() == ((b"\x01" * 4096) * 10) + ((b"\x00" * 4096) * 50)

    entry = fs.get("snapshot/sparse_end")
    assert entry.size == 0x3C000
    assert entry.extents() == [
        (0, 0, 0xD3C000, 20 * 4096, 0, 1 * 4096),
        (0, 0, 0xD5C000, 1 * 4096, 0, 1 * 4096),
        (0, 0, 0xD3C000, 20 * 4096, 2 * 4096, 18 * 4096),
        (0, 0, 0, 0, 0, 40 * 4096),
    ]
    assert entry.open().read() == (b"\x01" * (4096 + 123)) + (b"\x02") + (b"\x01" * ((4096 * 19) - 124)) + (
        (b"\x00" * 4096) * 40
    )



import stat

from dissect.cstruct import cstruct

btrfs_def = """
#define BTRFS_SUPER_INFO_OFFSET         0x00010000

/* ASCII for _BHRfS_M, no terminating nul */
#define BTRFS_MAGIC                     0x4D5F53665248425F

#define BTRFS_MAX_LEVEL                 8

/*
 * We can actually store much bigger names, but lets not confuse the rest of
 * linux.
 */
#define BTRFS_NAME_LEN                  255

/*
 * Theoretical limit is larger, but we keep this down to a sane value. That
 * should limit greatly the possibility of collisions on inode ref items.
 */
#define BTRFS_LINK_MAX                  65535

/* holds pointers to all of the tree roots */
#define BTRFS_ROOT_TREE_OBJECTID        1

/* stores information about which extents are in use, and reference counts */
#define BTRFS_EXTENT_TREE_OBJECTID      2

/*
 * chunk tree stores translations from logical -> physical block numbering
 * the super block points to the chunk tree
 */
#define BTRFS_CHUNK_TREE_OBJECTID       3

/*
 * stores information about which areas of a given device are in use.
 * one per device.  The tree of tree roots points to the device tree
 */
#define BTRFS_DEV_TREE_OBJECTID         4

/* one per subvolume, storing files and directories */
#define BTRFS_FS_TREE_OBJECTID          5

/* directory objectid inside the root tree */
#define BTRFS_ROOT_TREE_DIR_OBJECTID    6

/* holds checksums of all the data extents */
#define BTRFS_CSUM_TREE_OBJECTID        7

/* holds quota configuration and tracking */
#define BTRFS_QUOTA_TREE_OBJECTID       8

/* for storing items that use the BTRFS_UUID_KEY* types */
#define BTRFS_UUID_TREE_OBJECTID        9

/* tracks free space in block groups. */
#define BTRFS_FREE_SPACE_TREE_OBJECTID  10

/* Holds the block group items for extent tree v2. */
#define BTRFS_BLOCK_GROUP_TREE_OBJECTID 11

/* device stats in the device tree */
#define BTRFS_DEV_STATS_OBJECTID        0

/* for storing balance parameters in the root tree */
#define BTRFS_BALANCE_OBJECTID          -4

/* orphan objectid for tracking unlinked/truncated files */
#define BTRFS_ORPHAN_OBJECTID           -5

/* does write ahead logging to speed up fsyncs */
#define BTRFS_TREE_LOG_OBJECTID         -6
#define BTRFS_TREE_LOG_FIXUP_OBJECTID   -7

/* for space balancing */
#define BTRFS_TREE_RELOC_OBJECTID       -8
#define BTRFS_DATA_RELOC_TREE_OBJECTID  -9

/*
 * extent checksums all have this objectid
 * this allows them to share the logging tree
 * for fsyncs
 */
#define BTRFS_EXTENT_CSUM_OBJECTID      -10

/* For storing free space cache */
#define BTRFS_FREE_SPACE_OBJECTID       -11

/*
 * The inode number assigned to the special inode for storing
 * free ino cache
 */
#define BTRFS_FREE_INO_OBJECTID         -12

/* dummy objectid represents multiple objectids */
#define BTRFS_MULTIPLE_OBJECTIDS        -255

/*
 * All files have objectids in this range.
 */
#define BTRFS_FIRST_FREE_OBJECTID       256
#define BTRFS_LAST_FREE_OBJECTID        -256
#define BTRFS_FIRST_CHUNK_TREE_OBJECTID 256

/*
 * the device items go into the chunk tree.  The key is in the form
 * [ 1 BTRFS_DEV_ITEM_KEY device_id ]
 */
#define BTRFS_DEV_ITEMS_OBJECTID        1

#define BTRFS_BTREE_INODE_OBJECTID      1

#define BTRFS_EMPTY_SUBVOL_DIR_OBJECTID 2

#define BTRFS_DEV_REPLACE_DEVID         0

/*
 * inode items have the data typically returned from stat and store other
 * info about object characteristics.  There is one for every file and dir in
 * the FS
 */
#define BTRFS_INODE_ITEM_KEY            1
#define BTRFS_INODE_REF_KEY             12
#define BTRFS_INODE_EXTREF_KEY          13
#define BTRFS_XATTR_ITEM_KEY            24

/*
 * fs verity items are stored under two different key types on disk.
 * The descriptor items:
 * [ inode objectid, BTRFS_VERITY_DESC_ITEM_KEY, offset ]
 *
 * At offset 0, we store a btrfs_verity_descriptor_item which tracks the size
 * of the descriptor item and some extra data for encryption.
 * Starting at offset 1, these hold the generic fs verity descriptor.  The
 * latter are opaque to btrfs, we just read and write them as a blob for the
 * higher level verity code.  The most common descriptor size is 256 bytes.
 *
 * The merkle tree items:
 * [ inode objectid, BTRFS_VERITY_MERKLE_ITEM_KEY, offset ]
 *
 * These also start at offset 0, and correspond to the merkle tree bytes.  When
 * fsverity asks for page 0 of the merkle tree, we pull up one page starting at
 * offset 0 for this key type.  These are also opaque to btrfs, we're blindly
 * storing whatever fsverity sends down.
 */
#define BTRFS_VERITY_DESC_ITEM_KEY      36
#define BTRFS_VERITY_MERKLE_ITEM_KEY    37

#define BTRFS_ORPHAN_ITEM_KEY           48
/* reserve 2-15 close to the inode for later flexibility */

/*
 * dir items are the name -> inode pointers in a directory.  There is one
 * for every name in a directory.  BTRFS_DIR_LOG_ITEM_KEY is no longer used
 * but it's still defined here for documentation purposes and to help avoid
 * having its numerical value reused in the future.
 */
#define BTRFS_DIR_LOG_ITEM_KEY          60
#define BTRFS_DIR_LOG_INDEX_KEY         72
#define BTRFS_DIR_ITEM_KEY              84
#define BTRFS_DIR_INDEX_KEY             96
/*
 * extent data is for file data
 */
#define BTRFS_EXTENT_DATA_KEY           108

/*
 * extent csums are stored in a separate tree and hold csums for
 * an entire extent on disk.
 */
#define BTRFS_EXTENT_CSUM_KEY           128

/*
 * root items point to tree roots.  They are typically in the root
 * tree used by the super block to find all the other trees
 */
#define BTRFS_ROOT_ITEM_KEY             132

/*
 * root backrefs tie subvols and snapshots to the directory entries that
 * reference them
 */
#define BTRFS_ROOT_BACKREF_KEY          144

/*
 * root refs make a fast index for listing all of the snapshots and
 * subvolumes referenced by a given root.  They point directly to the
 * directory item in the root that references the subvol
 */
#define BTRFS_ROOT_REF_KEY              156

/*
 * extent items are in the extent map tree.  These record which blocks
 * are used, and how many references there are to each block
 */
#define BTRFS_EXTENT_ITEM_KEY           168

/*
 * The same as the BTRFS_EXTENT_ITEM_KEY, except it's metadata we already know
 * the length, so we save the level in key->offset instead of the length.
 */
#define BTRFS_METADATA_ITEM_KEY         169

#define BTRFS_TREE_BLOCK_REF_KEY        176

#define BTRFS_EXTENT_DATA_REF_KEY       178

#define BTRFS_EXTENT_REF_V0_KEY         180

#define BTRFS_SHARED_BLOCK_REF_KEY      182

#define BTRFS_SHARED_DATA_REF_KEY       184

/*
 * block groups give us hints into the extent allocation trees.  Which
 * blocks are free etc etc
 */
#define BTRFS_BLOCK_GROUP_ITEM_KEY      192

/*
 * Every block group is represented in the free space tree by a free space info
 * item, which stores some accounting information. It is keyed on
 * (block_group_start, FREE_SPACE_INFO, block_group_length).
 */
#define BTRFS_FREE_SPACE_INFO_KEY       198

/*
 * A free space extent tracks an extent of space that is free in a block group.
 * It is keyed on (start, FREE_SPACE_EXTENT, length).
 */
#define BTRFS_FREE_SPACE_EXTENT_KEY     199

/*
 * When a block group becomes very fragmented, we convert it to use bitmaps
 * instead of extents. A free space bitmap is keyed on
 * (start, FREE_SPACE_BITMAP, length); the corresponding item is a bitmap with
 * (length / sectorsize) bits.
 */
#define BTRFS_FREE_SPACE_BITMAP_KEY     200

#define BTRFS_DEV_EXTENT_KEY            204
#define BTRFS_DEV_ITEM_KEY              216
#define BTRFS_CHUNK_ITEM_KEY            228

/*
 * Records the overall state of the qgroups.
 * There's only one instance of this key present,
 * (0, BTRFS_QGROUP_STATUS_KEY, 0)
 */
#define BTRFS_QGROUP_STATUS_KEY         240
/*
 * Records the currently used space of the qgroup.
 * One key per qgroup, (0, BTRFS_QGROUP_INFO_KEY, qgroupid).
 */
#define BTRFS_QGROUP_INFO_KEY           242
/*
 * Contains the user configured limits for the qgroup.
 * One key per qgroup, (0, BTRFS_QGROUP_LIMIT_KEY, qgroupid).
 */
#define BTRFS_QGROUP_LIMIT_KEY          244
/*
 * Records the child-parent relationship of qgroups. For
 * each relation, 2 keys are present:
 * (childid, BTRFS_QGROUP_RELATION_KEY, parentid)
 * (parentid, BTRFS_QGROUP_RELATION_KEY, childid)
 */
#define BTRFS_QGROUP_RELATION_KEY       246

/*
 * Obsolete name, see BTRFS_TEMPORARY_ITEM_KEY.
 */
#define BTRFS_BALANCE_ITEM_KEY          248

/*
 * The key type for tree items that are stored persistently, but do not need to
 * exist for extended period of time. The items can exist in any tree.
 *
 * [subtype, BTRFS_TEMPORARY_ITEM_KEY, data]
 *
 * Existing items:
 *
 * - balance status item
 *   (BTRFS_BALANCE_OBJECTID, BTRFS_TEMPORARY_ITEM_KEY, 0)
 */
#define BTRFS_TEMPORARY_ITEM_KEY        248

/*
 * Obsolete name, see BTRFS_PERSISTENT_ITEM_KEY
 */
#define BTRFS_DEV_STATS_KEY             249

/*
 * The key type for tree items that are stored persistently and usually exist
 * for a long period, eg. filesystem lifetime. The item kinds can be status
 * information, stats or preference values. The item can exist in any tree.
 *
 * [subtype, BTRFS_PERSISTENT_ITEM_KEY, data]
 *
 * Existing items:
 *
 * - device statistics, store IO stats in the device tree, one key for all
 *   stats
 *   (BTRFS_DEV_STATS_OBJECTID, BTRFS_DEV_STATS_KEY, 0)
 */
#define BTRFS_PERSISTENT_ITEM_KEY       249

/*
 * Persistently stores the device replace state in the device tree.
 * The key is built like this: (0, BTRFS_DEV_REPLACE_KEY, 0).
 */
#define BTRFS_DEV_REPLACE_KEY           250

/*
 * Stores items that allow to quickly map UUIDs to something else.
 * These items are part of the filesystem UUID tree.
 * The key is built like this:
 * (UUID_upper_64_bits, BTRFS_UUID_KEY*, UUID_lower_64_bits).
 */
#define BTRFS_UUID_KEY_SUBVOL           251     /* for UUIDs assigned to subvols */
#define BTRFS_UUID_KEY_RECEIVED_SUBVOL  252     /* for UUIDs assigned to received subvols */

/*
 * string items are for debugging.  They just store a short string of
 * data in the FS
 */
#define BTRFS_STRING_ITEM_KEY           253

/* Maximum metadata block size (nodesize) */
#define BTRFS_MAX_METADATA_BLOCKSIZE    65536

/* 32 bytes in various csum fields */
#define BTRFS_CSUM_SIZE                 32

/*
 * flags definitions for directory entry item type
 *
 * Used by:
 * struct btrfs_dir_item.type
 *
 * Values 0..7 must match common file type values in fs_types.h.
 */
#define BTRFS_FT_UNKNOWN                0
#define BTRFS_FT_REG_FILE               1
#define BTRFS_FT_DIR                    2
#define BTRFS_FT_CHRDEV                 3
#define BTRFS_FT_BLKDEV                 4
#define BTRFS_FT_FIFO                   5
#define BTRFS_FT_SOCK                   6
#define BTRFS_FT_SYMLINK                7
#define BTRFS_FT_XATTR                  8
#define BTRFS_FT_MAX                    9
/* Directory contains encrypted data */
#define BTRFS_FT_ENCRYPTED              0x80

/*
 * Inode flags
 */
#define BTRFS_INODE_NODATASUM           (1 << 0)
#define BTRFS_INODE_NODATACOW           (1 << 1)
#define BTRFS_INODE_READONLY            (1 << 2)
#define BTRFS_INODE_NOCOMPRESS          (1 << 3)
#define BTRFS_INODE_PREALLOC            (1 << 4)
#define BTRFS_INODE_SYNC                (1 << 5)
#define BTRFS_INODE_IMMUTABLE           (1 << 6)
#define BTRFS_INODE_APPEND              (1 << 7)
#define BTRFS_INODE_NODUMP              (1 << 8)
#define BTRFS_INODE_NOATIME             (1 << 9)
#define BTRFS_INODE_DIRSYNC             (1 << 10)
#define BTRFS_INODE_COMPRESS            (1 << 11)

#define BTRFS_INODE_ROOT_ITEM_INIT      (1 << 31)

#define BTRFS_VOL_NAME_MAX              255
#define BTRFS_LABEL_SIZE                256

#define BTRFS_FSID_SIZE                 16
#define BTRFS_UUID_SIZE                 16

/* different types of block groups (and chunks) */
flag BTRFS_BLOCK_GROUP : uint64 {
    DATA        = 0x0001
    SYSTEM      = 0x0002
    METADATA    = 0x0004
    RAID0       = 0x0008
    RAID1       = 0x0010
    DUP         = 0x0020
    RAID10      = 0x0040
    RAID5       = 0x0080
    RAID6       = 0x0100
    RAID1C3     = 0x0200
    RAID1C4     = 0x0400
};

/*
 * The key defines the order in the tree, and so it also defines (optimal)
 * block layout.
 *
 * objectid corresponds to the inode number.
 *
 * type tells us things about the object, and is a kind of stream selector.
 * so for a given inode, keys with type of 1 might refer to the inode data,
 * type of 2 may point to file data in the btree and type == 3 may point to
 * extents.
 *
 * offset is the starting byte offset for this key in the stream.
 *
 * btrfs_disk_key is in disk byte order.  struct btrfs_key is always
 * in cpu native order.  Otherwise they are identical and their sizes
 * should be the same (ie both packed)
 */
struct btrfs_disk_key {
    uint64      objectid;
    uint8       type;
    uint64      offset;
};

/*
 * Every tree block (leaf or node) starts with this header.
 */
struct btrfs_header {
    /* These first four must match the super block */
    char        csum[BTRFS_CSUM_SIZE];
    /* FS specific uuid */
    char        fsid[BTRFS_FSID_SIZE];
    /* Which block this node is supposed to live in */
    uint64      bytenr;
    uint64      flags;

    /* Allowed to be different from the super from here on down */
    char        chunk_tree_uuid[BTRFS_UUID_SIZE];
    uint64      generation;
    uint64      owner;
    uint32      nritems;
    uint8       level;
};

/*
 * This is a very generous portion of the super block, giving us room to
 * translate 14 chunks with 3 stripes each.
 */
#define BTRFS_SYSTEM_CHUNK_ARRAY_SIZE   2048

/*
 * Just in case we somehow lose the roots and are not able to mount, we store
 * an array of the roots from previous transactions in the super.
 */
#define BTRFS_NUM_BACKUP_ROOTS 4
struct btrfs_root_backup {
    uint64      tree_root;
    uint64      tree_root_gen;

    uint64      chunk_root;
    uint64      chunk_root_gen;

    uint64      extent_root;
    uint64      extent_root_gen;

    uint64      fs_root;
    uint64      fs_root_gen;

    uint64      dev_root;
    uint64      dev_root_gen;

    uint64      csum_root;
    uint64      csum_root_gen;

    uint64      total_bytes;
    uint64      bytes_used;
    uint64      num_devices;
    /* future */
    uint64      unused_64[4];

    uint8       tree_root_level;
    uint8       chunk_root_level;
    uint8       extent_root_level;
    uint8       fs_root_level;
    uint8       dev_root_level;
    uint8       csum_root_level;
    /* future and to align */
    char        unused_8[10];
};

/*
 * A leaf is full of items. offset and size tell us where to find the item in
 * the leaf (relative to the start of the data area)
 */
struct btrfs_item {
    struct btrfs_disk_key   key;
    uint32      offset;
    uint32      size;
};

/*
 * Leaves have an item area and a data area:
 * [item0, item1....itemN] [free space] [dataN...data1, data0]
 *
 * The data is separate from the items to get the keys closer together during
 * searches.
 */
struct btrfs_leaf {
    struct btrfs_header     header;
    struct btrfs_item       items[];
};

/*
 * All non-leaf blocks are nodes, they hold only keys and pointers to other
 * blocks.
 */
struct btrfs_key_ptr {
    struct btrfs_disk_key   key;
    uint64      blockptr;
    uint64      generation;
};

struct btrfs_node {
    struct btrfs_header     header;
    struct btrfs_key_ptr    ptrs[];
};

struct btrfs_dev_item {
    /* the internal btrfs device id */
    uint64      devid;

    /* size of the device */
    uint64      total_bytes;

    /* bytes used */
    uint64      bytes_used;

    /* optimal io alignment for this device */
    uint32      io_align;

    /* optimal io width for this device */
    uint32      io_width;

    /* minimal io size for this device */
    uint32      sector_size;

    /* type and info about this device */
    uint64      type;

    /* expected generation for this device */
    uint64      generation;

    /*
     * starting byte of this partition on the device,
     * to allow for stripe alignment in the future
     */
    uint64      start_offset;

    /* grouping information for allocation decisions */
    uint32      dev_group;

    /* seek speed 0-100 where 100 is fastest */
    uint8       seek_speed;

    /* bandwidth 0-100 where 100 is fastest */
    uint8       bandwidth;

    /* btrfs generated uuid for this device */
    char        uuid[BTRFS_UUID_SIZE];

    /* uuid of FS who owns this device */
    char        fsid[BTRFS_UUID_SIZE];
};

struct btrfs_stripe {
    uint64      devid;
    uint64      offset;
    char        dev_uuid[BTRFS_UUID_SIZE];
};

struct btrfs_chunk {
    /* size of this chunk in bytes */
    uint64      length;

    /* objectid of the root referencing this chunk */
    uint64      owner;

    uint64      stripe_len;
    BTRFS_BLOCK_GROUP   type;

    /* optimal io alignment for this chunk */
    uint32      io_align;

    /* optimal io width for this chunk */
    uint32      io_width;

    /* minimal io size for this chunk */
    uint32      sector_size;

    /* 2^16 stripes is quite a lot, a second limit is the size of a single
     * item in the btree
     */
    uint16      num_stripes;

    /* sub stripes only matter for raid10 */
    uint16      sub_stripes;
    struct btrfs_stripe     stripe[num_stripes];
    /* additional stripes go here */
};

/*
 * The super block basically lists the main trees of the FS.
 */
struct btrfs_super_block {
    /* The first 4 fields must match struct btrfs_header */
    char        csum[BTRFS_CSUM_SIZE];
    /* FS specific UUID, visible to user */
    char        fsid[BTRFS_FSID_SIZE];
    /* This block number */
    uint64      bytenr;
    uint64      flags;

    /* Allowed to be different from the btrfs_header from here own down */
    uint64      magic;
    uint64      generation;
    uint64      root;
    uint64      chunk_root;
    uint64      log_root;

    /*
     * This member has never been utilized since the very beginning, thus
     * it's always 0 regardless of kernel version.  We always use
     * generation + 1 to read log tree root.  So here we mark it deprecated.
     */
    uint64      __unused_log_root_transid;
    uint64      total_bytes;
    uint64      bytes_used;
    uint64      root_dir_objectid;
    uint64      num_devices;
    uint32      sectorsize;
    uint32      nodesize;
    uint32      __unused_leafsize;
    uint32      stripesize;
    uint32      sys_chunk_array_size;
    uint64      chunk_root_generation;
    uint64      compat_flags;
    uint64      compat_ro_flags;
    uint64      incompat_flags;
    uint16      csum_type;
    uint8       root_level;
    uint8       chunk_root_level;
    uint8       log_root_level;
    struct btrfs_dev_item       dev_item;

    char        label[BTRFS_LABEL_SIZE];

    uint64      cache_generation;
    uint64      uuid_tree_generation;

    /* The UUID written into btree blocks */
    char        metadata_uuid[BTRFS_FSID_SIZE];

    uint64      nr_global_roots;

    /* Future expansion */
    uint64      reserved[27];
    char        sys_chunk_array[BTRFS_SYSTEM_CHUNK_ARRAY_SIZE];
    struct btrfs_root_backup    super_roots[BTRFS_NUM_BACKUP_ROOTS];

    /* Padded to 4096 bytes */
    char        padding[565];
};

struct btrfs_inode_ref {
    uint64      index;
    uint16      name_len;
    char        name[name_len];
};

struct btrfs_inode_extref {
    uint64      parent_objectid;
    uint64      index;
    uint16      name_len;
    char        name[name_len];
};

struct btrfs_timespec {
    uint64      sec;
    uint32      nsec;
};

struct btrfs_inode_item {
    /* nfs style generation number */
    uint64      generation;
    /* transid that last touched this inode */
    uint64      transid;
    uint64      size;
    uint64      nbytes;
    uint64      block_group;
    uint32      nlink;
    uint32      uid;
    uint32      gid;
    uint32      mode;
    uint64      rdev;
    uint64      flags;

    /* modification sequence number for NFS */
    uint64      sequence;

    /*
     * a little future expansion, for more than this we can
     * just grow the inode item and version it
     */
    uint64      reserved[4];
    struct btrfs_timespec   atime;
    struct btrfs_timespec   ctime;
    struct btrfs_timespec   mtime;
    struct btrfs_timespec   otime;
};

struct btrfs_dir_item {
    struct btrfs_disk_key   location;
    uint64      transid;
    uint16      data_len;
    uint16      name_len;
    uint8       type;
    char        name[name_len];
    char        data[data_len];
};

struct btrfs_root_item {
    struct btrfs_inode_item     inode;
    uint64      generation;
    uint64      root_dirid;
    uint64      bytenr;
    uint64      byte_limit;
    uint64      bytes_used;
    uint64      last_snapshot;
    uint64      flags;
    uint32      refs;
    struct btrfs_disk_key   drop_progress;
    uint8       drop_level;
    uint8       level;

    /*
     * The following fields appear after subvol_uuids+subvol_times
     * were introduced.
     */

    /*
     * This generation number is used to test if the new fields are valid
     * and up to date while reading the root item. Every time the root item
     * is written out, the "generation" field is copied into this field. If
     * anyone ever mounted the fs with an older kernel, we will have
     * mismatching generation values here and thus must invalidate the
     * new fields. See btrfs_update_root and btrfs_find_last_root for
     * details.
     * the offset of generation_v2 is also used as the start for the memset
     * when invalidating the fields.
     */
    uint64      generation_v2;
    char        uuid[BTRFS_UUID_SIZE];
    char        parent_uuid[BTRFS_UUID_SIZE];
    char        received_uuid[BTRFS_UUID_SIZE];
    uint64      ctransid; /* updated when an inode changes */
    uint64      otransid; /* trans when created */
    uint64      stransid; /* trans when sent. non-zero for received subvol */
    uint64      rtransid; /* trans when received. non-zero for received subvol */
    struct btrfs_timespec   ctime;
    struct btrfs_timespec   otime;
    struct btrfs_timespec   stime;
    struct btrfs_timespec   rtime;
    uint64 reserved[8]; /* for future */
};

/*
 * this is used for both forward and backward root refs
 */
struct btrfs_root_ref {
    uint64      dirid;
    uint64      sequence;
    uint16      name_len;
    char        name[name_len];
};

enum {
    BTRFS_FILE_EXTENT_INLINE   = 0,
    BTRFS_FILE_EXTENT_REG      = 1,
    BTRFS_FILE_EXTENT_PREALLOC = 2,
    BTRFS_NR_FILE_EXTENT_TYPES = 3,
};

enum {
    BTRFS_COMPRESS_NONE  = 0,
    BTRFS_COMPRESS_ZLIB  = 1,
    BTRFS_COMPRESS_LZO   = 2,
    BTRFS_COMPRESS_ZSTD  = 3,
    BTRFS_NR_COMPRESS_TYPES = 4,
};

struct btrfs_file_extent_item_inline {
    /*
     * transaction id that created this extent
     */
    uint64      generation;
    /*
     * max number of bytes to hold this extent in ram
     * when we split a compressed extent we can't know how big
     * each of the resulting pieces will be.  So, this is
     * an upper limit on the size of the extent in ram instead of
     * an exact limit.
     */
    uint64      ram_bytes;

    /*
     * 32 bits for the various ways we might encode the data,
     * including compression and encryption.  If any of these
     * are set to something a given disk format doesn't understand
     * it is treated like an incompat flag for reading and writing,
     * but not for stat.
     */
    uint8       compression;
    uint8       encryption;
    uint16      other_encoding; /* spare for later use */

    /* are we inline data or a real extent? */
    uint8       type;
};

struct btrfs_file_extent_item_reg {
    /*
     * transaction id that created this extent
     */
    uint64      generation;
    /*
     * max number of bytes to hold this extent in ram
     * when we split a compressed extent we can't know how big
     * each of the resulting pieces will be.  So, this is
     * an upper limit on the size of the extent in ram instead of
     * an exact limit.
     */
    uint64      ram_bytes;

    /*
     * 32 bits for the various ways we might encode the data,
     * including compression and encryption.  If any of these
     * are set to something a given disk format doesn't understand
     * it is treated like an incompat flag for reading and writing,
     * but not for stat.
     */
    uint8       compression;
    uint8       encryption;
    uint16      other_encoding; /* spare for later use */

    /* are we inline data or a real extent? */
    uint8       type;

    /*
     * disk space consumed by the extent, checksum blocks are included
     * in these numbers
     *
     * At this offset in the structure, the inline extent data start.
     */
    uint64      disk_bytenr;
    uint64      disk_num_bytes;
    /*
     * the logical offset in file blocks (no csums)
     * this extent record is for.  This allows a file extent to point
     * into the middle of an existing extent on disk, sharing it
     * between two snapshots (useful if some bytes in the middle of the
     * extent have changed
     */
    uint64      offset;
    /*
     * the logical number of file blocks (no csums included).  This
     * always reflects the size uncompressed and without encoding.
     */
    uint64      num_bytes;
};
"""

c_btrfs = cstruct().load(btrfs_def)

BTRFS_BLOCK_GROUP = c_btrfs.BTRFS_BLOCK_GROUP

BTRFS_BLOCK_GROUP_TYPE_MASK = BTRFS_BLOCK_GROUP.DATA | BTRFS_BLOCK_GROUP.SYSTEM | BTRFS_BLOCK_GROUP.METADATA

BTRFS_BLOCK_GROUP_PROFILE_MASK = (
    BTRFS_BLOCK_GROUP.RAID0
    | BTRFS_BLOCK_GROUP.RAID1
    | BTRFS_BLOCK_GROUP.RAID1C3
    | BTRFS_BLOCK_GROUP.RAID1C4
    | BTRFS_BLOCK_GROUP.RAID5
    | BTRFS_BLOCK_GROUP.RAID6
    | BTRFS_BLOCK_GROUP.DUP
    | BTRFS_BLOCK_GROUP.RAID10
)
BTRFS_BLOCK_GROUP_RAID56_MASK = BTRFS_BLOCK_GROUP.RAID5 | BTRFS_BLOCK_GROUP.RAID6

BTRFS_BLOCK_GROUP_RAID1_MASK = BTRFS_BLOCK_GROUP.RAID1 | BTRFS_BLOCK_GROUP.RAID1C3 | BTRFS_BLOCK_GROUP.RAID1C4

BTRFS_BLOCK_GROUP_STRIPE_MASK = BTRFS_BLOCK_GROUP.RAID0 | BTRFS_BLOCK_GROUP.RAID10 | BTRFS_BLOCK_GROUP_RAID56_MASK

FT_MAP = {
    c_btrfs.BTRFS_FT_UNKNOWN: None,
    c_btrfs.BTRFS_FT_REG_FILE: stat.S_IFREG,
    c_btrfs.BTRFS_FT_DIR: stat.S_IFDIR,
    c_btrfs.BTRFS_FT_CHRDEV: stat.S_IFCHR,
    c_btrfs.BTRFS_FT_BLKDEV: stat.S_IFBLK,
    c_btrfs.BTRFS_FT_FIFO: stat.S_IFIFO,
    c_btrfs.BTRFS_FT_SOCK: stat.S_IFSOCK,
    c_btrfs.BTRFS_FT_SYMLINK: stat.S_IFLNK,
}

BTRFS_RAID_ATTRIBUTES = {
    # (ncopies, nparity, tolerated_failures)
    BTRFS_BLOCK_GROUP(0): (1, 0, 0),  # BTRFS_RAID_SINGLE
    BTRFS_BLOCK_GROUP.RAID0: (1, 0, 0),
    BTRFS_BLOCK_GROUP.RAID1: (2, 0, 1),
    BTRFS_BLOCK_GROUP.DUP: (2, 0, 0),
    BTRFS_BLOCK_GROUP.RAID10: (2, 0, 1),
    BTRFS_BLOCK_GROUP.RAID5: (1, 1, 1),
    BTRFS_BLOCK_GROUP.RAID6: (1, 2, 2),
    BTRFS_BLOCK_GROUP.RAID1C3: (3, 0, 2),
    BTRFS_BLOCK_GROUP.RAID1C4: (4, 0, 3),
}



from __future__ import annotations

import io
import zlib
from bisect import bisect_right
from typing import TYPE_CHECKING, BinaryIO, NamedTuple
from uuid import UUID

from dissect.util.compression import lzo
from dissect.util.stream import AlignedStream

try:
    import zstandard

    HAS_ZSTD = True
except ImportError:
    HAS_ZSTD = False

from dissect.btrfs.c_btrfs import (
    BTRFS_BLOCK_GROUP,
    BTRFS_BLOCK_GROUP_PROFILE_MASK,
    BTRFS_BLOCK_GROUP_RAID1_MASK,
    BTRFS_BLOCK_GROUP_RAID56_MASK,
    BTRFS_BLOCK_GROUP_STRIPE_MASK,
    BTRFS_RAID_ATTRIBUTES,
    c_btrfs,
)
from dissect.btrfs.exceptions import Error

if TYPE_CHECKING:
    from dissect.btrfs.btrfs import Btrfs


class Stripe(NamedTuple):
    fh: BinaryIO
    offset: int
    devid: int
    dev_uuid: UUID


class Chunk(NamedTuple):
    offset: int
    length: int
    stripe_length: int
    type: c_btrfs.BTRFS_BLOCK_GROUP
    num_stripes: int
    sub_stripes: int
    data_stripes: int
    stripes: list[Stripe]


class ChunkStream(AlignedStream):
    """Implements a stream over Btrfs chunks, including basic RAID support.

    Args:
        btrfs: The filesystem this stream belongs to.
    """

    def __init__(self, btrfs: Btrfs):
        self.btrfs = btrfs
        self.chunks: list[Chunk] = []
        self._chunk_offsets: list[int] = []
        super().__init__(align=0x10000)

    def add(self, offset: int, chunk: c_btrfs.btrfs_chunk) -> None:
        """Add a chunk to the stream.

        This will iterate all stripes and link them to the appropriate devices.
        Allows for missing devices, if the RAID profile allows this.

        Args:
            offset: The logical offset to add this chunk for.
            chunk: The chunk item to add.
        """

        chunk_idx = bisect_right(self._chunk_offsets, offset)
        if chunk_idx > 0:
            # Check if we already have a chunk for this offset
            existing_chunk = self.chunks[chunk_idx - 1]
            if existing_chunk.offset <= offset and existing_chunk.offset + existing_chunk.length > offset:
                return

        ncopies, nparity, tolerated_failures = BTRFS_RAID_ATTRIBUTES[chunk.type & BTRFS_BLOCK_GROUP_PROFILE_MASK]
        data_stripes = (chunk.num_stripes - nparity) // ncopies

        stripes = []
        missing_devices = 0
        for stripe in chunk.stripe:
            fh = self.btrfs.devices.get(stripe.devid)
            dev_uuid = UUID(bytes=stripe.dev_uuid)

            if fh is None:
                if missing_devices < tolerated_failures:
                    missing_devices += 1
                else:
                    raise Error(f"Missing stripe disk for chunk offset {offset:#x}: {stripe.devid} ({dev_uuid})")

            stripe = Stripe(fh, stripe.offset, stripe.devid, dev_uuid)
            stripes.append(stripe)

        chunk = Chunk(
            offset,
            chunk.length,
            chunk.stripe_len,
            chunk.type,
            chunk.num_stripes,
            chunk.sub_stripes,
            data_stripes,
            stripes,
        )

        idx = bisect_right(self._chunk_offsets, offset)
        self._chunk_offsets.insert(idx, offset)
        self.chunks.insert(idx, chunk)

    def _read(self, offset: int, length: int) -> bytes:
        r = []

        chunk_idx = bisect_right(self._chunk_offsets, offset)
        chunks_len = len(self.chunks)

        while length > 0:
            if chunk_idx > chunks_len:
                # We somehow requested more data than we have chunks for
                break

            if chunk_idx == 0:
                # Read below the lowest chunk, just fill with zero bytes until the next chunk
                read_count = min(self._chunk_offsets[0] - offset, length)
                r.append(b"\x00" * read_count)

                offset += read_count
                length -= read_count
            else:
                chunk = self.chunks[chunk_idx - 1]

                chunk_offset = offset - chunk.offset
                chunk_remaining = chunk.length - chunk_offset

                while length > 0 and chunk_remaining > 0:
                    stripe_num, stripe_idx, stripe_offset, stripe_remaining = _get_stripe_read_info(chunk, chunk_offset)
                    stripe_read = min(stripe_remaining, length)

                    stripe = chunk.stripes[stripe_idx % chunk.num_stripes]
                    while stripe.fh is None:
                        # We already checked for the maximum amount of tolerated failures when adding the chunk,
                        # so looping here should be safe
                        if chunk.type & BTRFS_BLOCK_GROUP.DUP:
                            stripe_idx = 1
                        elif chunk.type & BTRFS_BLOCK_GROUP_RAID56_MASK:
                            raise NotImplementedError("RAID56 recovery is not yet supported")
                        else:
                            stripe_idx += 1

                        stripe = chunk.stripes[stripe_idx % chunk.num_stripes]

                    stripe.fh.seek(stripe.offset + stripe_offset + stripe_num * chunk.stripe_length)
                    r.append(stripe.fh.read(stripe_read))

                    offset += stripe_read
                    length -= stripe_read
                    chunk_offset += stripe_read
                    chunk_remaining -= stripe_read

            chunk_idx += 1

        return b"".join(r)


def _get_stripe_read_info(chunk: Chunk, offset: int) -> tuple[int, int, int, int]:
    # Reference: __btrfs_map_block
    stripe_num, stripe_offset = divmod(offset, chunk.stripe_length)
    stripe_idx = 0

    if chunk.type & BTRFS_BLOCK_GROUP.RAID0:
        stripe_num, stripe_idx = divmod(stripe_num, chunk.num_stripes)
    elif chunk.type & BTRFS_BLOCK_GROUP_RAID1_MASK:
        # We don't care from which mirror we read
        stripe_idx = 0
    elif chunk.type & BTRFS_BLOCK_GROUP.DUP:
        # We don't care from which duplicate we read
        stripe_idx = 0
    elif chunk.type & BTRFS_BLOCK_GROUP.RAID10:
        factor = chunk.num_stripes // chunk.sub_stripes
        stripe_num, stripe_idx = divmod(stripe_num, factor)
    elif chunk.type & BTRFS_BLOCK_GROUP_RAID56_MASK:
        stripe_num, stripe_idx = divmod(stripe_num, chunk.data_stripes)
        stripe_idx = (stripe_num + stripe_idx) % chunk.num_stripes
    else:
        stripe_num, stripe_idx = divmod(stripe_num, chunk.num_stripes)

    if chunk.type & BTRFS_BLOCK_GROUP_STRIPE_MASK:
        stripe_remaining = chunk.stripe_length - stripe_offset
    else:
        stripe_remaining = chunk.length - offset

    return stripe_num, stripe_idx, stripe_offset, stripe_remaining


class Extent(NamedTuple):
    compression: int
    encryption: int
    disk_offset: int
    disk_length: int
    offset: int
    length: int


class ExtentStream(AlignedStream):
    """Implement a stream over Btrfs file extents.

    Supports compression.

    Args:
        fh: The file-like object to read the data from, usually the :class:`ChunkStream` of the filesystem.
        extents: The list of extents to open the stream on.
        sector_size: The sector size of the filesystem, necessary for LZO decompression.
    """

    def __init__(self, fh: BinaryIO, extents: list[Extent], size: int, sector_size: int):
        self._fh = fh

        self.extents = extents
        self._extent_offsets = []

        offset = 0
        for extent in self.extents:
            if offset != 0:
                self._extent_offsets.append(offset)
            offset += extent.length

        self.sector_size = sector_size

        super().__init__(size, sector_size)

    def _read(self, offset: int, length: int) -> bytes:
        result = []

        extent_idx = bisect_right(self._extent_offsets, offset)
        extents_len = len(self.extents)
        size = self.size

        while length > 0:
            if extent_idx >= extents_len:
                # We somehow requested more data than we have runs for
                break

            # If run_idx == 0, we only have a single run
            extent_pos = 0 if extent_idx == 0 else self._extent_offsets[extent_idx - 1]
            extent = self.extents[extent_idx]
            extent_pos = offset - extent_pos
            extent_remaining = extent.length - extent_pos

            # The relative extent offset is only relevant for knowing where to actually start reading on the disk
            # Add it to the current extent_pos because that's where this variable is only going to be used for
            extent_pos += extent.offset

            # Sometimes the self.size is way larger than what we actually have runs for?
            # Stop reading if we reach a negative run_remaining
            if extent_remaining < 0:
                break

            read_count = min(size - offset, min(extent_remaining, length))

            # Sparse run
            if (extent.disk_offset, extent.disk_length) == (0, 0):
                result.append(b"\x00" * read_count)
            else:
                if (extent.compression, extent.encryption) == (c_btrfs.BTRFS_COMPRESS_NONE, 0):
                    # Quick path for no compression and no encryption
                    self._fh.seek(extent.disk_offset + extent_pos)
                    buf = self._fh.read(read_count)
                else:
                    self._fh.seek(extent.disk_offset)
                    buf = decode_extent(
                        self._fh.read(extent.disk_length),
                        extent.compression,
                        extent.encryption,
                        self.sector_size,
                    )

                    if extent_pos or read_count != len(buf):
                        buf = buf[extent_pos : extent_pos + read_count]
                result.append(buf)

            offset += read_count
            length -= read_count
            extent_idx += 1

        return b"".join(result)


def decode_extent(buf: bytes, compression: int, encryption: int, sector_size: int) -> bytes:
    """Decode a compressed extent.

    Args:
        buf: The extent data to decode.
        compression: The compression type to decompress from.
        encryption: The encryption type - currently unused.
        sector_size: The sector size of the filesystem, necessary for LZO decompression.
    """
    if compression == c_btrfs.BTRFS_COMPRESS_ZLIB:
        buf = zlib.decompress(buf)
    elif compression == c_btrfs.BTRFS_COMPRESS_LZO:
        # Reference: lzo.c
        lzo_buf = io.BytesIO(buf)
        lzo_buf.read(4)  # total size of compressed data
        lzo_out_size = _lzo_worst_compress(sector_size)

        decompressed_buf = bytearray()
        while True:
            lzo_segment_size = int.from_bytes(lzo_buf.read(4), "little")
            if lzo_segment_size == 0:
                break

            lzo_payload = lzo_buf.read(lzo_segment_size)
            decompressed_buf.extend(lzo.decompress(lzo_payload, False, lzo_out_size))

            sector_remaining = sector_size - (lzo_buf.tell() % sector_size)
            if sector_remaining >= 4:
                continue
            lzo_buf.seek(sector_remaining, io.SEEK_CUR)

        buf = bytes(decompressed_buf)
    elif compression == c_btrfs.BTRFS_COMPRESS_ZSTD:
        if not HAS_ZSTD:
            raise RuntimeError("Install `zstandard` to read zstandard compressed files")
        buf = zstandard.decompress(buf)

    if encryption:
        # btrfs doesn't actually support extent encryption yet
        raise NotImplementedError("BTRFS extent encryption is not supported")

    return buf


def _lzo_worst_compress(size: int) -> int:
    return size + (size // 16) + 64 + 3



from dissect.btrfs.btrfs import Btrfs, INode, Subvolume
from dissect.btrfs.exceptions import (
    Error,
    FileNotFoundError,
    NotADirectoryError,
    NotASymlinkError,
)

__all__ = [
    "Btrfs",
    "Error",
    "FileNotFoundError",
    "INode",
    "NotADirectoryError",
    "NotASymlinkError",
    "Subvolume",
]



# References:
# - https://btrfs.readthedocs.io/en/latest/
# - https://github.com/torvalds/linux/tree/master/fs/btrfs
# - https://github.com/btrfs/btrfs-dev-docs
# - https://www.sciencedirect.com/science/article/pii/S1742287618301993
from __future__ import annotations

import io
import stat
from functools import cache, cached_property, lru_cache
from typing import TYPE_CHECKING, BinaryIO
from uuid import UUID

from dissect.util import ts
from dissect.util.stream import BufferedStream

try:
    import warnings

    # If the C extension is not available, google-crc32c will display a warning
    with warnings.catch_warnings():
        warnings.simplefilter("ignore")
        from google_crc32c import extend as crc32c
except ImportError:
    from dissect.util.hash.crc32c import update as crc32c

from dissect.btrfs.c_btrfs import FT_MAP, c_btrfs
from dissect.btrfs.exceptions import (
    Error,
    FileNotFoundError,
    NotADirectoryError,
    NotASymlinkError,
)
from dissect.btrfs.stream import ChunkStream, Extent, ExtentStream, decode_extent
from dissect.btrfs.tree import BTree

if TYPE_CHECKING:
    from collections.abc import Iterator
    from datetime import datetime


class Btrfs:
    """Btrfs filesystem implementation.

    This implementation supports most basic Btrfs features such as subvolumes, compression and (meta)data RAID.
    To open a RAID volume, simply pass all file-like objects that belong to the RAID set as a list.

    Args:
        fh: A file-like object for the volume to use for parsing Btrfs.
    """

    def __init__(self, fh: BinaryIO | list[BinaryIO]):
        self.fhs = fh if isinstance(fh, list) else [fh]

        if not self.fhs:
            raise ValueError("At least one file-like object is required")

        sb_fhs = []
        for fh in self.fhs:
            fh.seek(c_btrfs.BTRFS_SUPER_INFO_OFFSET)
            sb = c_btrfs.btrfs_super_block(fh)
            if sb.magic != c_btrfs.BTRFS_MAGIC:
                raise Error("Invalid btrfs superblock")
            sb_fhs.append((sb, fh))

        if len({sb.fsid for sb, _ in sb_fhs}) > 1:
            raise ValueError("Only file-like objects of the same filesystem UUID can be used")

        sb_fhs = sorted(sb_fhs, key=lambda item: item[0].generation, reverse=True)

        self.sb = sb_fhs[0][0]
        self.sector_size = self.sb.sectorsize
        self.node_size = self.sb.nodesize
        self.stripe_size = self.sb.stripesize

        self.label = self.sb.label.split(b"\x00")[0].decode()
        self.uuid = UUID(bytes=self.sb.fsid)
        self.metadata_uuid = UUID(bytes=self.sb.metadata_uuid)

        self.devices = {sb.dev_item.devid: fh for sb, fh in sb_fhs}

        self._open_tree = lru_cache(32)(self._open_tree)
        self.open_subvolume = lru_cache(16)(self.open_subvolume)

        self._logical_fh = ChunkStream(self)
        self._initialize_chunks()

        self._root_tree = BTree(self, root_offset=self.sb.root)

        self.fs_tree = self.open_subvolume(c_btrfs.BTRFS_FS_TREE_OBJECTID)

        self.default_subvolume = self._open_default_subvolume()
        self.root = self.default_subvolume.root

    def get(self, path: str | int, node: INode | None = None) -> INode:
        """Retrieve a Btrfs inode by path or inode number.

        Args:
            path: Filesystem path or inode number.
            node: Optional inode used for relative lookups.
        """
        return self.default_subvolume.get(path, node)

    def subvolumes(self) -> Iterator[Subvolume]:
        """Yield all subvolumes."""
        yield self.fs_tree

        search_ids = [c_btrfs.BTRFS_FS_TREE_OBJECTID]

        cursor = self._root_tree.cursor()
        for object_id in search_ids:
            for item, _ in cursor.iter(
                objectid=object_id, type=c_btrfs.BTRFS_ROOT_REF_KEY, offset=0, ignore_offset=True
            ):
                search_ids.append(item.key.offset)
                yield self.open_subvolume(item.key.offset)
            cursor.reset()

    def find_subvolume(self, path: str | None = None) -> Subvolume | None:
        """Find a subvolume by path.

        Args:
            path: The path of the subvolume to find.
        """
        for subvolume in self.subvolumes():
            if subvolume.path == path:
                return subvolume

        return None

    def open_subvolume(self, objectid: int, parent: INode | None = None) -> Subvolume:
        """Open a subvolume.

        Args:
            objectid: The objectid of the subvolume to open.
            parent: Optional parent node to attach to the root of the subvolume.
        """
        return Subvolume(self, objectid, parent)

    def _initialize_chunks(self) -> None:
        """Initialize the logical data stream by reading the system chunk array and traversing the chunk tree."""
        sys_chunk_array = io.BytesIO(memoryview(self.sb.sys_chunk_array)[: self.sb.sys_chunk_array_size])
        while True:
            key = c_btrfs.btrfs_disk_key(sys_chunk_array)
            if key.type != c_btrfs.BTRFS_CHUNK_ITEM_KEY:
                raise ValueError(f"Invalid item type in sys_chunk_array: {key}")

            chunk = c_btrfs.btrfs_chunk(sys_chunk_array)
            self._logical_fh.add(key.offset, chunk)

            if sys_chunk_array.tell() == self.sb.sys_chunk_array_size:
                break

        chunk_tree = BTree(self, root_offset=self.sb.chunk_root)
        for item, data in chunk_tree.cursor().iter(
            objectid=c_btrfs.BTRFS_FIRST_CHUNK_TREE_OBJECTID,
            type=c_btrfs.BTRFS_CHUNK_ITEM_KEY,
            offset=0,
            ignore_offset=True,
        ):
            chunk = c_btrfs.btrfs_chunk(data)
            self._logical_fh.add(item.key.offset, chunk)

    def _open_tree(self, objectid: int) -> BTree:
        """Open a tree by object ID.

        Args:
            objectid: The object ID of the tree to open.
        """
        _, data = self._root_tree.find(objectid, c_btrfs.BTRFS_ROOT_ITEM_KEY)
        root_item = c_btrfs.btrfs_root_item(data)
        return BTree(self, root_item)

    def _open_default_subvolume(self) -> Subvolume:
        """Find and open the subvolume that's configured as default."""
        _, data = self._root_tree.find(c_btrfs.BTRFS_ROOT_TREE_DIR_OBJECTID, c_btrfs.BTRFS_DIR_ITEM_KEY)
        root_dir_item = c_btrfs.btrfs_dir_item(data)
        return Subvolume(self, root_dir_item.location.objectid)

    def _read_node(self, address: int) -> bytes:
        """Helper method for reading a node.

        Args:
            address: The node address to read.
        """
        self._logical_fh.seek(address)
        return self._logical_fh.read(self.node_size)


class Subvolume:
    """Represent a Btrfs subvolume.

    Btrfs has support for multiple subvolumes. The default subvolume is the ``FS_TREE`` subvolume.
    Each subvolume has its own B-tree.

    Args:
        btrfs: The filesystem this subvolume belongs to.
        objectid: The object ID of the subvolume to open.
        parent: Optional parent node to attach to the root of the subvolume.
    """

    def __init__(self, btrfs: Btrfs, objectid: int, parent: INode | None = None):
        self.btrfs = btrfs
        self.objectid = objectid
        self.parent = parent

        self.inode = lru_cache(8192)(self.inode)
        self.resolve_path = lru_cache(1024)(self.resolve_path)

    def __repr__(self) -> str:
        return f"<Subvolume objectid={self.objectid}>"

    @cached_property
    def tree(self) -> BTree:
        return self.btrfs._open_tree(self.objectid)

    @cached_property
    def uuid(self) -> UUID:
        return UUID(bytes=self.tree.root_item.uuid)

    @cached_property
    def root(self) -> INode:
        return self.inode(self.tree.root_item.root_dirid, c_btrfs.BTRFS_FT_DIR, self.parent)

    @cached_property
    def path(self) -> str:
        parts = []
        objectid = self.objectid
        while objectid != c_btrfs.BTRFS_FS_TREE_OBJECTID:
            item, data = self.btrfs._root_tree.find(objectid=objectid, type=c_btrfs.BTRFS_ROOT_BACKREF_KEY)
            root_ref = c_btrfs.btrfs_root_ref(data)
            name = root_ref.name.decode(errors="surrogateescape")
            parts.append(name)

            if path := self.btrfs.open_subvolume(item.key.offset).resolve_path(root_ref.dirid):
                parts.append(path)

            objectid = item.key.offset

        return "/".join(parts[::-1])

    def get(self, path: str | int, node: INode | None = None) -> INode:
        """Retrieve a Btrfs inode by path or inode number.

        Args:
            path: Filesystem path or inode number.
            node: Optional inode used for relative lookups.
        """
        if isinstance(path, int):
            return self.inode(path)

        node = node or self.root
        subvolume = self

        parts = path.encode().split(b"/")

        for part in parts:
            if not part:
                continue

            if part == b".":
                continue

            if part == b"..":
                node = node.parent or node
                continue

            while node.is_symlink():
                node = node.link_inode

            # The Linux kernel doesn't do an initial and final XOR with 0xFFFFFFFF
            # Btrfs uses an initial CRC of `(u32)~1`, which is effectively the same as 1 XOR 0xFFFFFFFF
            # We still need to invert the XOR of the result though
            # https://stackoverflow.com/a/40433980
            part_hash = (crc32c(1, part) ^ 0xFFFFFFFF) & 0xFFFFFFFF
            try:
                _, data = subvolume.tree.find(node.inum, c_btrfs.BTRFS_DIR_ITEM_KEY, part_hash)
            except KeyError:
                raise FileNotFoundError(f"File not found: {path}")

            dir_item = c_btrfs.btrfs_dir_item(data)
            dir_item_type = dir_item.location.type

            if dir_item_type == c_btrfs.BTRFS_ROOT_ITEM_KEY:
                subvolume = self.btrfs.open_subvolume(dir_item.location.objectid, subvolume)
                node = subvolume.root
            elif dir_item_type == c_btrfs.BTRFS_INODE_ITEM_KEY:
                node = subvolume.inode(dir_item.location.objectid, dir_item.type, subvolume)
            else:
                raise NotImplementedError(f"Unknown dir_item type: {dir_item}")

        return node

    def inode(self, inum: int, type: int | None = None, parent: INode | None = None) -> INode:
        """Return an :class:`INode` by number, optionally attaching a type and parent."""
        return INode(self, inum, type, parent)

    def resolve_path(self, objectid: int) -> str:
        names = []
        while objectid != c_btrfs.BTRFS_FIRST_FREE_OBJECTID:
            item, data = self.tree.find(objectid, c_btrfs.BTRFS_INODE_REF_KEY)
            inode_ref = c_btrfs.btrfs_inode_ref(data)
            names.append(inode_ref.name.decode(errors="surrogateescape"))

            objectid = item.key.offset

        return "/".join(names[::-1])


class INode:
    """Represent a Btrfs inode.

    Args:
        subvolume: The subvolume this inode belongs to.
        inum: The inode number of this inode.
        type: Optional file type of this inode, as observed in a directory entry.
        parent: Optional parent of this inode, if this inode is parsed from a directory listing.
    """

    def __init__(self, subvolume: Subvolume, inum: int, type: int | None = None, parent: INode | None = None):
        self.subvolume = subvolume
        self.btrfs = subvolume.btrfs
        self.inum = inum
        self._type = type
        self.parent = parent

        self.listdir = cache(self.listdir)
        self.extents = cache(self.extents)

    def __repr__(self) -> str:
        return f"<inode {self.subvolume.objectid}:{self.inum}>"

    @cached_property
    def inode(self) -> c_btrfs.btrfs_inode_item:
        """Return the parsed inode structure."""
        _, data = self.subvolume.tree.find(self.inum, c_btrfs.BTRFS_INODE_ITEM_KEY)
        return c_btrfs.btrfs_inode_item(data)

    @cached_property
    def size(self) -> int:
        """Return the file size."""
        return self.inode.size

    @cached_property
    def uid(self) -> int:
        """Return the owner user ID."""
        return self.inode.uid

    @cached_property
    def gid(self) -> int:
        """Return the owner group ID."""
        return self.inode.gid

    @cached_property
    def mode(self) -> int:
        """Return the file mode."""
        return self.inode.mode

    @cached_property
    def type(self) -> int:
        """Return the file type."""
        return FT_MAP.get(self._type) or stat.S_IFMT(self.inode.mode)

    @cached_property
    def atime(self) -> datetime:
        """Return datetime timestamp of last access."""
        return ts.from_unix_ns(self.atime_ns)

    @cached_property
    def atime_ns(self) -> int:
        """Return nanosecond timestamp of last access."""
        return _parse_ts(self.inode.atime)

    @cached_property
    def ctime(self) -> datetime:
        """Return datetime timestamp of last metadata change."""
        return ts.from_unix_ns(self.ctime_ns)

    @cached_property
    def ctime_ns(self) -> int:
        """Return nanosecond timestamp of last metadata change."""
        return _parse_ts(self.inode.ctime)

    @cached_property
    def mtime(self) -> datetime:
        """Return datetime timestamp of last content modification."""
        return ts.from_unix_ns(self.mtime_ns)

    @cached_property
    def mtime_ns(self) -> int:
        """Return nanosecond timestamp of last content modification."""
        return _parse_ts(self.inode.mtime)

    @cached_property
    def otime(self) -> datetime:
        """Return datetime timestamp of inode creation."""
        return ts.from_unix_ns(self.otime_ns)

    @cached_property
    def otime_ns(self) -> int:
        """Return nanosecond timestamp of inode creation."""
        return _parse_ts(self.inode.otime)

    def is_dir(self) -> bool:
        """Return whether this inode is a directory."""
        return self.type == stat.S_IFDIR

    def is_file(self) -> bool:
        """Return whether this inode is a regular file."""
        return self.type == stat.S_IFREG

    def is_symlink(self) -> bool:
        """Return whether this inode is a symlink."""
        return self.type == stat.S_IFLNK

    def is_block_device(self) -> bool:
        """Return whether this inode is a block device."""
        return self.type == stat.S_IFBLK

    def is_character_device(self) -> bool:
        """Return whether this inode is a character device."""
        return self.type == stat.S_IFCHR

    def is_device(self) -> bool:
        """Return whether this inode is a device."""
        return self.is_block_device() or self.is_character_device()

    def is_fifo(self) -> bool:
        """Return whether this inode is a FIFO file."""
        return self.type == stat.S_IFIFO

    def is_socket(self) -> bool:
        """Return whether this inode is a socket file."""
        return self.type == stat.S_IFSOCK

    def is_ipc(self) -> bool:
        """Return whether this inode is an IPC file."""
        return self.is_fifo() or self.is_socket()

    @cached_property
    def link(self) -> str:
        """Return the symlink target."""
        if not self.is_symlink():
            raise NotASymlinkError(f"{self!r} is not a symlink")

        return self.open().read().decode(errors="surrogateescape")

    @cached_property
    def link_inode(self) -> INode:
        """Resolve the symlink target to an inode."""
        link = self.link
        relnode = None if link.startswith("/") else self.parent
        return self.subvolume.get(self.link, relnode)

    @property
    def parents(self) -> Iterator[INode]:
        for item, data in self.subvolume.tree.cursor().iter(
            self.inum, c_btrfs.BTRFS_INODE_REF_KEY, 0, ignore_offset=True
        ):
            inode_ref = c_btrfs.btrfs_inode_ref(data)
            if inode_ref.name == b"..":
                if self.parent:
                    yield self.parent
            else:
                yield self.subvolume.inode(item.key.offset, c_btrfs.BTRFS_FT_DIR)

    @property
    def path(self) -> str:
        """Return the path to this inode within the subvolume. In case of multiple hardlinks, return the first path."""
        return next(self.paths())

    @property
    def full_path(self) -> str:
        """Return the full path to this inode. In case of multiple hardlinks, return the first path."""
        return next(self.paths(True))

    def get(self, path: str) -> INode:
        """Retrieve a Btrfs inode relative from this inode.

        Args:
            path: Filesystem path.
        """
        return self.subvolume.get(path, self)

    def paths(self, full: bool = False) -> Iterator[str]:
        """Yield all paths (hardlinks) to this inode.

        By default only resolves up to the root of the subvolume this inode belongs to.
        For a full path to the root of the filesystem tree, set ``full`` to ``True``.

        Args:
            full: Whether to fully resolve the path up to the root of the filesystem tree.
        """
        root = self.subvolume.path if full else ""
        for item, data in self.subvolume.tree.cursor().iter(
            self.inum, c_btrfs.BTRFS_INODE_REF_KEY, 0, ignore_offset=True
        ):
            if item.key.offset == self.inum:
                yield root
                break

            inode_ref = c_btrfs.btrfs_inode_ref(data)
            name = inode_ref.name.decode(errors="surrogateescape")

            path = [name]
            if parent_path := self.subvolume.resolve_path(item.key.offset):
                path.append(parent_path)

            if root:
                path.append(root)

            yield "/".join(path[::-1])

    def listdir(self) -> dict[str, INode]:
        """Return a directory listing."""
        return dict(self.iterdir())

    def iterdir(self) -> Iterator[tuple[str, INode]]:
        """Iterate directory contents."""
        if not self.is_dir():
            raise NotADirectoryError(f"{self!r} is not a directory")

        yield ".", self
        yield "..", self.parent or self

        # The offset in BTRFS_DIR_INDEX_KEY items is the directory index
        # Start searching from index 2 (because . and .. are 0 and 1 respectively)
        cursor = self.subvolume.tree.cursor()
        for _, data in cursor.iter(self.inum, c_btrfs.BTRFS_DIR_INDEX_KEY, 2, ignore_offset=True):
            dir_item = c_btrfs.btrfs_dir_item(data)
            dir_item_type = dir_item.location.type
            name = dir_item.name.decode(errors="surrogateescape")

            if dir_item_type == c_btrfs.BTRFS_ROOT_ITEM_KEY:
                subvolume = self.btrfs.open_subvolume(dir_item.location.objectid, self)
                yield name, subvolume.root
            elif dir_item_type == c_btrfs.BTRFS_INODE_ITEM_KEY:
                yield name, self.subvolume.inode(dir_item.location.objectid, dir_item.type, self)
            else:
                raise NotImplementedError(f"Unknown dir_item type: {dir_item}")

    def extents(self) -> list[Extent] | None:
        with self.open() as fh:
            if isinstance(fh, ExtentStream):
                return fh.extents

        return None

    def open(self) -> BinaryIO:
        """Return the data stream for the inode.

        File data in Btrfs can be inlined in the B-tree or stored in file extents. In both cases it can be compressed.
        """
        if not self.size:
            return BufferedStream(io.BytesIO(), size=0)

        offset = 0
        extents = []

        cursor = self.subvolume.tree.cursor()
        for item, data in cursor.iter(self.inum, c_btrfs.BTRFS_EXTENT_DATA_KEY, 0, ignore_offset=True):
            extent = c_btrfs.btrfs_file_extent_item_inline(data)

            if extent.type == c_btrfs.BTRFS_FILE_EXTENT_INLINE:
                header_len = len(c_btrfs.btrfs_file_extent_item_inline)
                buf = decode_extent(
                    data[header_len:],
                    extent.compression,
                    extent.encryption,
                    self.btrfs.sector_size,
                )
                return BufferedStream(io.BytesIO(buf), size=self.size)

            extent = c_btrfs.btrfs_file_extent_item_reg(data)
            if extent.type == c_btrfs.BTRFS_FILE_EXTENT_REG:
                key = item.key

                if offset < key.offset:
                    gap = key.offset - offset
                    extents.append(Extent(0, 0, 0, 0, 0, gap))
                    offset += gap

                extents.append(
                    Extent(
                        extent.compression,
                        extent.encryption,
                        extent.disk_bytenr,
                        extent.disk_num_bytes,
                        extent.offset,
                        extent.num_bytes,
                    )
                )
                offset += extent.num_bytes

        if offset < self.size:
            extents.append(Extent(0, 0, 0, 0, 0, self.size - offset))

        return ExtentStream(self.btrfs._logical_fh, extents, self.size, self.btrfs.sector_size)


def _parse_ts(timespec: c_btrfs.btrfs_timespec) -> int:
    """Parse a Btrfs time specification into a nanosecond timestamp."""
    return (timespec.sec * 1000000000) + timespec.nsec
